/***************************************************************************
 *   Copyright (C) 2008 by Dominik Seichter, Craig Ringer                  *
 *   domseichter@web.de                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Library General Public License as       *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this program; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "../PdfDefines.h"
#include "../PdfDefinesPrivate.h"

#if defined(PODOFO_MULTI_THREAD)
#error "Multi-thread build, a real PdfMutex implementation should be used instead"
#endif

namespace PoDoFo {
namespace Util {

/**
 * A platform independent non-reentrant mutex, no-op implementation.
 * This version is used if PoDoFo is built without threading support.
 *  
 * PdfMutex is *NOT* part of PoDoFo's public API.
 */
class PdfMutexImpl {
  public:
    /** Construct a new mutex
     */
    inline PdfMutexImpl() { }

    inline ~PdfMutexImpl() { }

    /**
     * Lock the mutex
     */
    inline void Lock() { }

    /**
     * Try locking the mutex. 
     *
     * \returns true if the mutex was locked
     * \returns false if the mutex is already locked
     *                by some other thread
     */
    inline bool TryLock() { return true; }

    /**
     * Unlock the mutex
     */
    inline void UnLock() { }
};

}; // Util
}; // PoDoFo
